<?php

namespace Drupal\gutenberg;

/**
 * A.
 */
class TinyColor {

  /**
   * Direct port of tinycolor's bound01 function.
   *
   * It's lightly simplified to maintain consistency with tinycolor.
   *
   * @param mixed $n
   *   Number of unknown type.
   * @param int $max
   *   Upper value of the range to bound to.
   *
   * @see https://github.com/bgrins/TinyColor
   *
   * @return float
   *   Value in the range [0,1].
   */
  public static function bound01($n, $max) {
    if ('string' === gettype($n) && FALSE !== strpos($n, '.') && 1 === (float) $n) {
      $n = '100%';
    }

    $n = min($max, max(0, (float) $n));

    // Automatically convert percentage into number.
    if ('string' === gettype($n) && FALSE !== strpos($n, '%')) {
      $n = (int) ($n * $max) / 100;
    }

    // Handle floating point rounding errors.
    if ((abs($n - $max) < 0.000001)) {
      return 1.0;
    }

    // Convert into [0, 1] range if it isn't already.
    return ($n % $max) / (float) $max;
  }

  /**
   * Direct port of tinycolor's boundAlpha function.
   *
   * Direct port of tinycolor's boundAlpha function to maintain consistency
   * with how tinycolor works.
   *
   * @param mixed $n
   *   Number of unknown type.
   *
   * @see https://github.com/bgrins/TinyColor
   *
   * @return float
   *   Value in the range [0,1].
   */
  public function boundAlpha($n) {
    if (is_numeric($n)) {
      $n = (float) $n;
      if ($n >= 0 && $n <= 1) {
        return $n;
      }
    }
    return 1;
  }

  /**
   * Round and convert values of an RGB object.
   *
   * @param array $rgb_color
   *   RGB object.
   *
   * @see https://github.com/bgrins/TinyColor
   *
   * @return array
   *   Rounded and converted RGB object.
   */
  public function rgbToRgb(array $rgb_color) {
    return [
      'r' => $this->bound01($rgb_color['r'], 255) * 255,
      'g' => $this->bound01($rgb_color['g'], 255) * 255,
      'b' => $this->bound01($rgb_color['b'], 255) * 255,
    ];
  }

  /**
   * Helper function for hsl to rgb conversion.
   *
   * @param float $p
   *   First component.
   * @param float $q
   *   Second component.
   * @param float $t
   *   Third component.
   *
   * @see https://github.com/bgrins/TinyColor
   *
   * @return float
   *   R, G, or B component.
   */
  public function hueToRgb($p, $q, $t) {
    if ($t < 0) {
      $t += 1;
    }
    if ($t > 1) {
      $t -= 1;
    }
    if ($t < 1 / 6) {
      return $p + ($q - $p) * 6 * $t;
    }
    if ($t < 1 / 2) {
      return $q;
    }
    if ($t < 2 / 3) {
      return $p + ($q - $p) * (2 / 3 - $t) * 6;
    }
    return $p;
  }

  /**
   * Convert an HSL object to an RGB object with converted and rounded values.
   *
   * @param array $hsl_color
   *   HSL object.
   *
   * @see https://github.com/bgrins/TinyColor
   *
   * @return array
   *   Rounded and converted RGB object.
   */
  public function hslToRgb(array $hsl_color) {
    $h = $this->bound01($hsl_color['h'], 360);
    $s = $this->bound01($hsl_color['s'], 100);
    $l = $this->bound01($hsl_color['l'], 100);

    if (0 === $s) {
      // Achromatic.
      $r = $l;
      $g = $l;
      $b = $l;
    }
    else {
      $q = $l < 0.5 ? $l * (1 + $s) : $l + $s - $l * $s;
      $p = 2 * $l - $q;
      $r = $this->hueToRgb($p, $q, $h + 1 / 3);
      $g = $this->hueToRgb($p, $q, $h);
      $b = $this->hueToRgb($p, $q, $h - 1 / 3);
    }

    return [
      'r' => $r * 255,
      'g' => $g * 255,
      'b' => $b * 255,
    ];
  }

  /**
   * Parses hex, hsl, and rgb CSS strings.
   *
   * Parses hex, hsl, and rgb CSS strings. using the same regex as tinycolor
   * v1.4.2 used in the JavaScript. Only colors output from react-color are
   * implemented.
   *
   * @param string $color_str
   *   CSS color string.
   *
   * @see https://github.com/bgrins/TinyColor
   * @see https://github.com/casesandberg/react-color/
   *
   * @return array
   *   RGB object.
   */
  public function stringToRgb($color_str) {
    $color_str = strtolower(trim($color_str));

    $css_integer = '[-\\+]?\\d+%?';
    $css_number  = '[-\\+]?\\d*\\.\\d+%?';

    $css_unit = '(?:' . $css_number . ')|(?:' . $css_integer . ')';

    $permissive_match3 = '[\\s|\\(]+(' . $css_unit . ')[,|\\s]+(' . $css_unit . ')[,|\\s]+(' . $css_unit . ')\\s*\\)?';
    $permissive_match4 = '[\\s|\\(]+(' . $css_unit . ')[,|\\s]+(' . $css_unit . ')[,|\\s]+(' . $css_unit . ')[,|\\s]+(' . $css_unit . ')\\s*\\)?';

    $rgb_regexp = '/^rgb' . $permissive_match3 . '$/';
    if (preg_match($rgb_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => $match[1],
        'g' => $match[2],
        'b' => $match[3],
      ]
      );

      $rgb['a'] = 1;

      return $rgb;
    }

    $rgba_regexp = '/^rgba' . $permissive_match4 . '$/';
    if (preg_match($rgba_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => $match[1],
        'g' => $match[2],
        'b' => $match[3],
      ]
      );

      $rgb['a'] = $this->boundAlpha($match[4]);

      return $rgb;
    }

    $hsl_regexp = '/^hsl' . $permissive_match3 . '$/';
    if (preg_match($hsl_regexp, $color_str, $match)) {
      $rgb = $this->hslToRgb(
      [
        'h' => $match[1],
        's' => $match[2],
        'l' => $match[3],
      ]
      );

      $rgb['a'] = 1;

      return $rgb;
    }

    $hsla_regexp = '/^hsla' . $permissive_match4 . '$/';
    if (preg_match($hsla_regexp, $color_str, $match)) {
      $rgb = $this->hslToRgb(
      [
        'h' => $match[1],
        's' => $match[2],
        'l' => $match[3],
      ]
      );

      $rgb['a'] = $this->boundAlpha($match[4]);

      return $rgb;
    }

    $hex8_regexp = '/^#?([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/';
    if (preg_match($hex8_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => base_convert($match[1], 16, 10),
        'g' => base_convert($match[2], 16, 10),
        'b' => base_convert($match[3], 16, 10),
      ]
      );

      $rgb['a'] = $this->boundAlpha(
      base_convert($match[4], 16, 10) / 255
      );

      return $rgb;
    }

    $hex6_regexp = '/^#?([0-9a-fA-F]{2})([0-9a-fA-F]{2})([0-9a-fA-F]{2})$/';
    if (preg_match($hex6_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => base_convert($match[1], 16, 10),
        'g' => base_convert($match[2], 16, 10),
        'b' => base_convert($match[3], 16, 10),
      ]
      );

      $rgb['a'] = 1;

      return $rgb;
    }

    $hex4_regexp = '/^#?([0-9a-fA-F]{1})([0-9a-fA-F]{1})([0-9a-fA-F]{1})([0-9a-fA-F]{1})$/';
    if (preg_match($hex4_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => base_convert($match[1] . $match[1], 16, 10),
        'g' => base_convert($match[2] . $match[2], 16, 10),
        'b' => base_convert($match[3] . $match[3], 16, 10),
      ]
      );

      $rgb['a'] = $this->boundAlpha(
      base_convert($match[4] . $match[4], 16, 10) / 255
      );

      return $rgb;
    }

    $hex3_regexp = '/^#?([0-9a-fA-F]{1})([0-9a-fA-F]{1})([0-9a-fA-F]{1})$/';
    if (preg_match($hex3_regexp, $color_str, $match)) {
      $rgb = $this->rgbToRgb(
      [
        'r' => base_convert($match[1] . $match[1], 16, 10),
        'g' => base_convert($match[2] . $match[2], 16, 10),
        'b' => base_convert($match[3] . $match[3], 16, 10),
      ]
      );

      $rgb['a'] = 1;

      return $rgb;
    }

    // The JS color picker considers the string "transparent" to be a hex value,
    // so we need to handle it here as a special case.
    if ('transparent' === $color_str) {
      return [
        'r' => 0,
        'g' => 0,
        'b' => 0,
        'a' => 0,
      ];
    }
  }

  /**
   * Registers the style and colors block attributes.
   *
   * Registers the style and colors block attributes for block types that
   * support it.
   *
   * @param WP_Block_Type $block_type Block Type.
   */
  /* function gutenberg_register_duotone_support( $block_type ) {
   *  $has_duotone_support = false;
   *  if ( property_exists( $block_type, 'supports' ) ) {
   *    $has_duotone_support = _wp_array_get($block_type->supports, [
   *      'color', '__experimentalDuotone'
   *    ], false );
   *  }
   *
   *  if ( $has_duotone_support ) {
   *    if ( ! $block_type->attributes ) {
   *      $block_type->attributes = array();
   *    }
   *
   *    if ( ! array_key_exists( 'style', $block_type->attributes ) ) {
   *      $block_type->attributes['style'] = array(
   *        'type' => 'object',
   *      );
   *    }
   *  }
   * } */

  /**
   * Renders the duotone filter SVG and returns the CSS filter property.
   *
   * Renders the duotone filter SVG and returns the CSS filter property to
   * reference the rendered SVG.
   *
   * @param array $preset
   *   Duotone preset value as seen in theme.json.
   *
   * @return string
   *   Duotone CSS filter property.
   */
  public function gutenbergRenderDuotoneFilterPreset(array $preset) {
    $duotone_id     = $preset['slug'];
    $duotone_colors = $preset['colors'];
    $filter_id      = 'wp-duotone-' . $duotone_id;
    $duotone_values = [
      'r' => [],
      'g' => [],
      'b' => [],
      'a' => [],
    ];
    foreach ($duotone_colors as $color_str) {
      $color = gutenberg_tinycolor_string_to_rgb($color_str);

      $duotone_values['r'][] = $color['r'] / 255;
      $duotone_values['g'][] = $color['g'] / 255;
      $duotone_values['b'][] = $color['b'] / 255;
      $duotone_values['a'][] = $color['a'];
    }

    ob_start();

    ?>

        <svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 0 0" width="0" height="0" focusable="false" role="none" style="visibility: hidden; position: absolute; left: -9999px; overflow: hidden;">
            <defs>
                <filter id="<?php echo esc_attr($filter_id); ?>">
                    <feColorMatrix color-interpolation-filters="sRGB" type="matrix" values="
                        .299 .587 .114 0 0
                        .299 .587 .114 0 0
                        .299 .587 .114 0 0
                        .299 .587 .114 0 0
                    " />
                    <feComponentTransfer color-interpolation-filters="sRGB">
                        <feFuncR type="table" tableValues="<?php echo esc_attr(implode(' ', $duotone_values['r'])); ?>" />
                        <feFuncG type="table" tableValues="<?php echo esc_attr(implode(' ', $duotone_values['g'])); ?>" />
                        <feFuncB type="table" tableValues="<?php echo esc_attr(implode(' ', $duotone_values['b'])); ?>" />
                        <feFuncA type="table" tableValues="<?php echo esc_attr(implode(' ', $duotone_values['a'])); ?>" />
                    </feComponentTransfer>
                    <feComposite in2="SourceGraphic" operator="in" />
                </filter>
            </defs>
        </svg>

    <?php

    $svg = ob_get_clean();

    if (!defined('SCRIPT_DEBUG') || !SCRIPT_DEBUG) {
      // Clean up the whitespace.
      $svg = preg_replace("/[\r\n\t ]+/", ' ', $svg);
      $svg = preg_replace('/> </', '><', $svg);
      $svg = trim($svg);
    }

    /* add_action(
     *  // SVG filters won't render at all in the head of a document and
     *  // Safari incorrectly renders SVG filters in the footer, so the
     *  // beginning of the body seems to be the safest place to render.
     *  'wp_body_open',
     *  function () use ( $svg ) {
     *    echo $svg;
     *  }
     * ); */

    return "url('#" . $filter_id . "')";
  }

}
